#include <stdlib.h>
#include "Python.h"
#include "complexobject.h"

static PyObject *ErrorObject;

static PyObject *
orbiplot(PyObject *self, PyObject *args)
{
  PyObject *result;
  double c0,c1;
  char *image_string;
  int W, H, start, stop;

  if (!PyArg_ParseTuple(args, "iiiidd:iterate", &W, &H, &start, &stop, &c0, &c1))
    return NULL;
  if ( W <= 0 || H <= 0 ){
    PyErr_SetString( ErrorObject, "Image dimensions must be positive.");
    return NULL;
  }
  if ( c1 > 0.25 || c0 > c1){
    PyErr_SetString( ErrorObject, "This computation requires  c0 < c1 <= 0.25.");
    return NULL;
  }
  result = PyString_FromStringAndSize(NULL, W*H);
  if (result == NULL)
    return NULL;
  image_string = PyString_AsString(result);
  memset(image_string, '\377', W*H);
  {
    double delta_c = (c1 - c0)/W;
    //    double p = (1 + sqrt(1-4*c0))/2;
    double p = 2;
    double delta_x = 2*p/H; 
    double x;
    double c = c0;
    int i, j, pix;
    for (i=0; i<W; i++){
      x = 0;
      for (j=0; j<start; j++)
	x = x*x+c;
      for (j=start; j<stop; j++) {
	x = x*x+c;
	pix = (int)((p-x)/delta_x);
	if (0 <= pix && pix < H)
	  image_string[i+pix*W] = '\0';
      }
      c += delta_c;
    }
  }
  return result;
}   

static char orbi_doc[]=
"\n"
"Generates an image (as a string) by iterating x -> x^2 + c.\n"
"\n"
"Arguments:\n"
"   W       integer        image width\n"
"   H       integer        image height\n"
"   start   integer        first iterate to display\n"
"   stop     integer       last iterate to display\n"
"   c0, c1  floats         endpoints of c-interval; c0 < c1 <= .25\n"
"\n"
"The interval [c0,c1] is sampled at W points.  For each value of c\n"
"the tail end of the orbit of 0 under x -> x^2 + c is plotted on a\n"
"vertical line.  Orbit points are plotted if they lie between p and -p\n"
"where p = (1+sqrt(1-4*c0))/2\n";


static PyObject *ErrorObject;

static PyObject *
iterate(PyObject *self, PyObject *args)
{
  PyObject *result;
  Py_complex z0,z1,c0,c1;
  char *image_string;
  int W, H, max;

  if (!PyArg_ParseTuple(args, "iiiDDDD:iterate", &W, &H, &max, &z0, &z1, &c0, &c1))
    return NULL;
  if ( W <= 0 || H <= 0 ){
    PyErr_SetString( ErrorObject, "Image dimensions must be positive.");
    return NULL;
  }
  if ( max < 0 || max > 255 ){
    PyErr_SetString( ErrorObject, "The value of max must lie between 0 and 255."); 
    return NULL;
  }
  result = PyString_FromStringAndSize(NULL, W*H);
  if (result == NULL)
    return NULL;
  image_string = PyString_AsString(result);

  {
    double *z_real = (double *)malloc(W*sizeof(double));
    double *z_imag = (double *)malloc(H*sizeof(double));
    double *c_real = (double *)malloc(W*sizeof(double));
    double *c_imag = (double *)malloc(H*sizeof(double));
    double delta_z, delta_c,z,c;
    register double R, I, RR, II, mod;
    int i, j;
    int iterations;

    if (z_real == NULL || z_imag == NULL || c_real == NULL || c_imag == NULL)
      return NULL;
    z = z0.real;
    c = c0.real;
    delta_z = (z1.real - z)/W;
    delta_c = (c1.real - c)/W;
    for (i=0; i<W; i++){
      z_real[i] = z;
      z += delta_z;
      c_real[i] = c;
      c += delta_c;
    }

    z = z1.imag;
    c = c1.imag;
    delta_z = (z0.imag - z)/H;
    delta_c = (c0.imag - c)/H;
    for (j=0; j<H; j++){
      z_imag[j] = z;
      z += delta_z;
      c_imag[j] = c;
      c += delta_c;
    }

    for (j=0; j<H; j++){
      for (i=0; i<W; i++){
	I = z_imag[j];
	R = z_real[i];
	iterations = 0;
	while (iterations < max){
	  RR = R*R;
	  II = I*I;
	  mod = RR + II;
	  if (mod > 4) break;
	  I = 2*I*R + c_imag[j];
	  R = RR - II + c_real[i];
	  ++iterations;
	}
	*image_string = iterations;
	++image_string;
      }
    }
    free(z_real);
    free(z_imag);
    free(c_real);
    free(c_imag);
  }
  return result;
}   

static char iterate_doc[]=
"Generates an image (as a string) by iterating z -> z^2 + c.\n"
"\n"
"Arguments:\n"
"   W       integer             image width\n"
"   H       integer             image height\n"
"   max     integer             maximum number of iterates (0 <= max <= 255)\n"
"   z0, z1  complex numbers     corners of a box in the z-domain\n"
"   c0, c1  complex numbers     corners of a box in the c-domain\n"
"\n"
"The boxes in the z- and c- domains are sampled on a grid of size WxH.\n"
"For each pair (i,j), the (i,j) grid point in the z-domain is used as an\n"
"initial z-value and the (i,j) grid point in the c-domain is used\n"
"as the value of c.  The function z -> z^2 + c is iterated up to max\n"
"times, stopping if |z| > 2.  The (i,j) pixel in the image is set\n"
"to the number of iterations performed.\n";

static PyObject *
boundary(PyObject *self, PyObject *args)
{
  PyObject *result;
  unsigned char *in_string;
  unsigned char *out_string;
  int W, H, max, length;
  int i, j, p;

  if (!PyArg_ParseTuple(args, "s#iii:boundary", &in_string, &length, &W, &H, &max))
    return NULL;
  if ( W <= 0 || H <= 0 || W*H != length){
    PyErr_SetString( ErrorObject, "Invalid image dimensions.");
    return NULL;
  }
  if ( max < 0 || max > 255 ){
    PyErr_SetString( ErrorObject, "The value of max must lie between 0 and 255."); 
    return NULL;
  }
  result = PyString_FromStringAndSize(NULL, W*H);
  if (result == NULL)
    return NULL;
  out_string = PyString_AsString(result);

  for (i=0; i<W; i++){
    for (j=0; j<H; j++){
      p = i+j*W;
      out_string[p] = 255;
      if (in_string[p] < max)
	if ( (i > 0 && in_string[p-1] >= max) ||
             (i < W && in_string[p+1] >= max) ||
             (j > 0 && in_string[p-W] >= max) ||
	     (j < H && in_string[p+W] >= max) ) out_string[p] = 0;
    }
  }
  return result;
}

static char boundary_doc[] =
"Generates a black and white image of the boundary of a colored region.\n"
"\n"
"Arguments:\n"
"    image   string     the image as a string\n"
"    W       integer    the width of the image\n"
"    H       integer    the height of the image\n"
"    max     integer    threshold\n"
"\n"
"Returns an image which is the boundary of the region in which pixel\n"
"values are less than max.\n";

static PyObject *
boxcount(PyObject *self, PyObject *args)
{
  PyObject *result;
  unsigned char *in_string;
  int W, H, max, length;
  int i, j;

  if (!PyArg_ParseTuple(args, "s#iii:boxcount",
			&in_string, &length, &W, &H, &max))
    return NULL;

  if ( W <= 0 || H <= 0 || W*H != length){
    PyErr_SetString( ErrorObject, "Invalid image dimensions.");
    return NULL;
  }

  if ( max < 0 || max > 255 ){
    PyErr_SetString( ErrorObject, "The value of max must lie between 0 and 255."); 
    return NULL;
  }

  result = PyList_New(0);
  if (result == NULL)
    return NULL;

  {
    unsigned char *buf = (unsigned char *)malloc(length*sizeof(unsigned char));
    int m, n, k;
    long count;
    PyObject *item;

    if (buf == NULL)
      return NULL;
    for (i = 0; i < length; i++){
      if (in_string[i] < max)
	buf[i] = '\001';
      else
	buf[i] = '\0';
    }
    k = (W < H ? W : H)>>2;
    while (k > 1){
      i=0;
      j=0;
      for (m=H ; m>0 ; m-=2){
	for (n=W ; n>0 ; n-=2){
	  if (n>1){
	    buf[i] = buf[j] | buf[j+1];
	    if (m>1)
	      buf[i] |= (buf[j+W] | buf[j+W+1]);
	    j += 2;
	  }
	  else {
	    buf[i] = buf[j];
	    if (m>1)
	      buf[i] |= buf[j+W];
	    j += 1;
	  }
	  ++i;
	}
	j += W;
      }
      W = (W+1)/2;
      H = (H+1)/2;
      count = 0;
      while (i > 0){
	count += buf[--i];
      }
      item = PyInt_FromLong(count);
      if (item == NULL)
        return NULL;
      if (PyList_Append(result, item) == -1)
        return NULL;
      Py_DECREF(item);

      k = k >> 1;
    }
    free(buf);
  }
  return result;
}


static char boxcount_doc[] =
"Inputs an image as a string. Counts how many boxes of size 2^i meet the image.\n"
"\n"
"Arguments:\n"
"    image   string     the image as a string\n"
"    W       integer    the width of the image\n"
"    H       integer    the height of the image\n"
"    max     integer    threshold\n"
"\n"
"A box is deemed to meet the image if it contains a pixel of value >= max.\n"
"Returns a list of integers, where the ith integer is the number of boxes of\n"
"size 2^i that meet the image.  The maximum box size is the largest power of 2\n"
"which is smaller than min(W,H)/4.\n";

/* List of functions defined in the module */

static PyMethodDef quaditer_methods[] = {
	{"iterate",	iterate,      METH_VARARGS, iterate_doc  },
	{"boundary",    boundary,     METH_VARARGS, boundary_doc },
	{"boxcount",    boxcount,     METH_VARARGS, boxcount_doc },
	{NULL,		NULL}		/* sentinel */
};


/* Initialization function for the module */

DL_EXPORT(void)
initquaditer(void)
{
	PyObject *m, *d;

	/* Create the module and add the functions */
	m = Py_InitModule("quaditer", quaditer_methods);

	/* Add some symbolic constants to the module */
	d = PyModule_GetDict(m);
	ErrorObject = PyErr_NewException("quaditer.error", NULL, NULL);
	PyDict_SetItemString(d, "error", ErrorObject);
}

/* List of functions defined in the module */

static PyMethodDef chaos_methods[] = {
	{"orbiplot",	orbiplot,     METH_VARARGS, orbi_doc },
	{"iterate",	iterate,      METH_VARARGS, iterate_doc  },
	{"boundary",    boundary,     METH_VARARGS, boundary_doc },
	{"boxcount",    boxcount,     METH_VARARGS, boxcount_doc },
	{NULL,		NULL}		/* sentinel */
};


/* Initialization function for the module */

DL_EXPORT(void)
init_chaos(void)
{
	PyObject *m, *d;

	/* Create the module and add the functions */
	m = Py_InitModule("_chaos", chaos_methods);

	/* Add some symbolic constants to the module */
	d = PyModule_GetDict(m);
	ErrorObject = PyErr_NewException("orbi.error", NULL, NULL);
	PyDict_SetItemString(d, "error", ErrorObject);
}
